#ifndef FrameCPP_VERSION_6_FrAdcData_HH
#define FrameCPP_VERSION_6_FrAdcData_HH

#if defined(__cplusplus)
#if ! defined(SWIGIMPORTED)
#include <memory>
#include <vector>
#endif /* ! defined(SWIGIMPORTED) */

#include "general/types.hh"

#include "framecpp/Common/Container.hh"
#include "framecpp/Common/TOCInfo.hh"

#include "framecpp/Version6/FrameSpec.hh"
#include "framecpp/Version6/Dimension.hh"
#include "framecpp/Version6/FrVect.hh"
#include "framecpp/Version6/STRING.hh"

#include "framecpp/Version4/FrAdcData.hh"
#endif /* defined(__cplusplus) */

#undef FR_ADC_DATA_CHANNEL_GROUP_TYPE
#define FR_ADC_DATA_CHANNEL_GROUP_TYPE	INT_4U
#undef FR_ADC_DATA_CHANNEL_NUMBER_TYPE
#define FR_ADC_DATA_CHANNEL_NUMBER_TYPE	INT_4U
#undef FR_ADC_DATA_N_BITS_TYPE
#define FR_ADC_DATA_N_BITS_TYPE		INT_4U
#undef FR_ADC_DATA_BIAS_TYPE
#define FR_ADC_DATA_BIAS_TYPE		REAL_4
#undef FR_ADC_DATA_SLOPE_TYPE
#define FR_ADC_DATA_SLOPE_TYPE		REAL_4
#undef FR_ADC_DATA_SAMPLERATE_TYPE
#define FR_ADC_DATA_SAMPLERATE_TYPE	REAL_8
#undef FR_ADC_DATA_TIMEOFFSET_TYPE
#define FR_ADC_DATA_TIMEOFFSET_TYPE	REAL_8
#undef FR_ADC_DATA_FSHIFT_TYPE
#define FR_ADC_DATA_FSHIFT_TYPE		REAL_8
#undef FR_ADC_DATA_PHASE_TYPE
#define FR_ADC_DATA_PHASE_TYPE		REAL_4
#undef FR_ADC_DATA_DATAVALID_TYPE
#define FR_ADC_DATA_DATAVALID_TYPE	INT_2U

#if defined(__cplusplus) && !defined(SWIGIMPORTED)
namespace FrameCPP
{
  namespace Version_6
  {
    const FR_ADC_DATA_BIAS_TYPE FR_ADC_DATA_DEFAULT_BIAS = 0.0;
    const FR_ADC_DATA_SLOPE_TYPE FR_ADC_DATA_DEFAULT_SLOPE = 1.0;
    const FR_ADC_DATA_FSHIFT_TYPE FR_ADC_DATA_DEFAULT_FSHIFT = 0.0;
    const FR_ADC_DATA_TIMEOFFSET_TYPE FR_ADC_DATA_DEFAULT_TIME_OFFSET = 0.0;
    const FR_ADC_DATA_DATAVALID_TYPE FR_ADC_DATA_DEFAULT_DATA_VALID = 0;
    const FR_ADC_DATA_PHASE_TYPE FR_ADC_DATA_DEFAULT_PHASE = 0.0;
    inline const char* FR_ADC_DATA_DEFAULT_UNITS( )
    {
      return "counts";
    }

    class FrAdcData
      : public Common::FrameSpec::Object,
	public Common::TOCInfo
    {
    public:
      typedef Common::Container< FrVect >	data_type;
      typedef data_type::const_iterator const_iterator;
      typedef data_type::iterator iterator;

      typedef Common::Container< FrVect >	aux_type;
      typedef aux_type::const_iterator const_aux_iterator;
      typedef aux_type::iterator aux_iterator;

      static const REAL_4 DEFAULT_BIAS;
      static const REAL_4 DEFAULT_SLOPE;
      static const REAL_8 DEFAULT_FSHIFT;
      static const REAL_8 DEFAULT_TIME_OFFSET;
      static const INT_2U DEFAULT_DATA_VALID;
      static const REAL_4 DEFAULT_PHASE;
      inline static const char* DEFAULT_UNITS( )
      {
	return FR_ADC_DATA_DEFAULT_UNITS( );
      }

      //: Default constructor
      FrAdcData( );

      FrAdcData( const FrAdcData& Source );

      FrAdcData( const std::string& name, INT_4U group,
		 INT_4U channel, INT_4U nBits,
		 REAL_8 sampleRate,
		 REAL_4 bias = DEFAULT_BIAS,
		 REAL_4 slope = DEFAULT_SLOPE,
		 const std::string& units = DEFAULT_UNITS( ),
		 REAL_8 fShift = DEFAULT_FSHIFT,
		 REAL_8 timeOffset = DEFAULT_TIME_OFFSET,
		 INT_2U dataValid = DEFAULT_DATA_VALID,
		 REAL_4 phase = DEFAULT_PHASE );

      FrAdcData( Previous::FrAdcData& Source, istream_type* Stream );

      ~FrAdcData( );

      static const char* StructName( );

      static const Common::Description* StructDescription( );

      //: Comparison operator
      bool operator==( const FrAdcData& RHS ) const;
      bool operator!=( const FrAdcData& RHS ) const;

      /// \brief comparison operator
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      //: Number of bytes needed to write this structure
      cmn_streamsize_type Bytes( const Common::StreamBase& Stream) const;

      virtual const char* ObjectStructName( ) const;

      // Accessors
      const std::string& GetName() const; 
      const std::string& GetComment( ) const;
      INT_4U GetChannelNumber( ) const;
      INT_4U GetChannelGroup( ) const;
      INT_4U GetNBits( ) const;
      REAL_4 GetBias( ) const;
      REAL_4 GetSlope( ) const;
      STRING GetUnits( ) const;
      REAL_8 GetSampleRate( ) const;
      REAL_8 GetTimeOffset( ) const;
      REAL_8 GetFShift( ) const;
      REAL_4 GetPhase( ) const;
      INT_2U GetDataValid( ) const;

      const aux_type& RefAux( ) const;
      aux_type& RefAux( );
      const data_type& RefData( ) const;
      data_type& RefData( );
   
      // Mutators
      void AppendComment( const std::string& comment );

      //: Merge with another FrAdcData
      FrAdcData& Merge( const FrAdcData& RHS );

      static promote_ret_type
      Promote( INT_2U Source,
	       promote_arg_type Obj,
	       istream_type* Stream );

      void SetDataValid( INT_2U flag );
      void SetBias( REAL_4 bias );
      void SetSlope( REAL_4 slope );
      void SetTimeOffset( REAL_8 time );

      std::auto_ptr< FrAdcData >
      SubFrAdcData( REAL_8 Offset, REAL_8 Dt ) const;
   
      virtual FrAdcData* Create( istream_type& Stream ) const;

      //: Write the structure to the stream
      virtual void Write( ostream_type& Stream ) const;

      virtual void TOCQuery( int InfoClass, ... ) const;

    protected:
      /// \brief Down grade an object
      virtual demote_ret_type
      demote( INT_2U Target,
	      demote_arg_type Obj,
	      istream_type* Stream ) const;

      /// \brief Upgrade an object
      virtual promote_ret_type
      promote( INT_2U Target,
	       promote_arg_type Obj,
	       istream_type* Stream ) const;

    private:
      using Common::FrameSpec::Object::Create;

      struct data_definition_type {
	STRING	name;
	STRING	comment;
	INT_4U	channelGroup;
	INT_4U	channelNumber;
	INT_4U	nBits;
	REAL_4	bias;
	REAL_4	slope;
	STRING	units;
	REAL_8	sampleRate;
	REAL_8	timeOffset;
	REAL_8	fShift;
	REAL_4	phase;
	INT_2U	dataValid;
	data_type	data;
	aux_type	aux;

	data_definition_type( );
	data_definition_type( const data_definition_type& Source );

	//: Comparison operator
	bool operator==( const data_definition_type& RHS ) const;
	bool operator!=( const data_definition_type& RHS ) const;

	void copy_core( const data_definition_type& Source );
      } m_data;

      FrAdcData( istream_type& Stream );
    };

    inline FrAdcData::promote_ret_type FrAdcData::
    Promote( INT_2U Source,
	     promote_arg_type Obj,
	     istream_type* Stream )
    {
      return Object::PromoteObject< Previous::FrAdcData, FrAdcData >
	( DATA_FORMAT_VERSION, Source, Obj, Stream );
    }

    inline const char* FrAdcData::
    StructName( )
    {
      static const char* name = "FrAdcData";
      return name;
    }

    inline const std::string& FrAdcData::
    GetComment( ) const { return m_data.comment; }

    inline INT_4U FrAdcData::
    GetChannelNumber( ) const { return m_data.channelNumber; }

    inline INT_4U FrAdcData::
    GetChannelGroup( ) const { return m_data.channelGroup; }

    inline INT_4U FrAdcData::
    GetNBits( ) const { return m_data.nBits; }

    inline REAL_4 FrAdcData::
    GetBias( ) const { return m_data.bias; }

    inline REAL_4 FrAdcData::
    GetSlope( ) const { return m_data.slope; }

    inline STRING FrAdcData::
    GetUnits( ) const { return m_data.units; }

    inline REAL_8 FrAdcData::
    GetSampleRate( ) const { return m_data.sampleRate; }

    inline REAL_8 FrAdcData::
    GetTimeOffset( ) const { return m_data.timeOffset; }

    inline REAL_8 FrAdcData::
    GetFShift( ) const { return m_data.fShift; }

    inline REAL_4 FrAdcData::
    GetPhase( ) const { return m_data.phase; }

    inline INT_2U FrAdcData::
    GetDataValid( ) const { return m_data.dataValid; }

    inline const FrAdcData::aux_type& FrAdcData::
    RefAux( ) const { return m_data.aux; }

    inline FrAdcData::aux_type& FrAdcData::
    RefAux( ) { return m_data.aux; }

    inline const FrAdcData::data_type& FrAdcData::
    RefData( ) const { return m_data.data; }

    inline FrAdcData::data_type& FrAdcData::
    RefData( ) { return m_data.data; }

    inline void FrAdcData::
    SetDataValid( INT_2U flag )
    {
      m_data.dataValid = flag;
    }

    inline void FrAdcData::
    SetBias( REAL_4 bias )
    {
      m_data.bias = bias;
    }

    inline void FrAdcData::
    SetSlope( REAL_4 slope )
    {
      m_data.slope = slope;
    }

    inline void FrAdcData::
    SetTimeOffset( REAL_8 time )
    {
      m_data.timeOffset = time;
    }

    inline bool FrAdcData::data_definition_type::
    operator!=( const data_definition_type& RHS ) const
    {
      return ! ( *this == RHS );
    }

    inline void FrAdcData::
    AppendComment( const std::string& comment )
    {
      FrameCPP::Common::AppendComment( m_data.comment, comment );
    } 

    inline bool FrAdcData::
    operator==( const FrAdcData& RHS ) const
    {
      return ( this->m_data == RHS.m_data );
    }

    inline bool FrAdcData::
    operator!=( const FrAdcData& RHS ) const
    {
      return ! ( *this == RHS );
    }
  } // namespace - Version_6
} // namespace FrameCPP
#endif /* defined(__cplusplus) && !defined(SWIGIMPORTED) */
#endif /* FrameCPP_VERSION_6_FrAdcData_HH */
