#ifndef FrameCPP_VERSION_8_FrProcData_HH
#define FrameCPP_VERSION_8_FrProcData_HH

#if defined(__cplusplus)
#ifndef SWIGIMPORTED
#include <memory>
#include <vector>
#endif /* SWIGIMPORTED */
#endif /* defined(__cplusplus) */

#include "general/types.hh"

#if defined(__cplusplus)
#ifndef SWIGIMPORTED
#include "framecpp/Common/Array.hh"
#include "framecpp/Common/Container.hh"
#include "framecpp/Common/TOCInfo.hh"

#include "framecpp/Version7/FrProcData.hh"

#include "framecpp/Version8/FrameSpec.hh"
#include "framecpp/Version8/Dimension.hh"
#include "framecpp/Version8/FrHistory.hh"
#include "framecpp/Version8/FrTable.hh"
#include "framecpp/Version8/FrVect.hh"
#include "framecpp/Version8/STRING.hh"
#endif /* SWIGIMPORTED */
#endif /* defined(__cplusplus) */

#undef	FR_PROC_DATA_TYPE_TYPE
#define FR_PROC_DATA_TYPE_TYPE		INT_2U
#undef	FR_PROC_DATA_SUB_TYPE_TYPE
#define	FR_PROC_DATA_SUB_TYPE_TYPE	INT_2U
#undef	FR_PROC_DATA_TIME_OFFSET_TYPE
#define	FR_PROC_DATA_TIME_OFFSET_TYPE	REAL_8
#undef	FR_PROC_DATA_T_RANGE_TYPE
#define	FR_PROC_DATA_T_RANGE_TYPE	REAL_8
#undef	FR_PROC_DATA_F_SHIFT_TYPE
#define	FR_PROC_DATA_F_SHIFT_TYPE	REAL_8
#undef	FR_PROC_DATA_PHASE_TYPE
#define	FR_PROC_DATA_PHASE_TYPE		REAL_4
#undef	FR_PROC_DATA_F_RANGE_TYPE
#define	FR_PROC_DATA_F_RANGE_TYPE	REAL_8
#undef	FR_PROC_DATA_BW_TYPE
#define	FR_PROC_DATA_BW_TYPE		REAL_8

#if defined(__cplusplus)
namespace FrameCPP
{
  namespace Version_8
  {
    //===================================================================
    // Data elements for an FrProcData
    //===================================================================
#if ! defined(SWIG)
    typedef Previous::FrProcData::FrProcDataData FrProcDataData;
#endif /* ! defined(SWIG) */
    //===================================================================
    /// \brief Post-Processed Data Structure Definition.
    //===================================================================
    class FrProcData
#if ! defined(SWIG)
      :	public ObjectWithChecksum< Common::FrameSpec::Object >,
	public FrProcDataData,
	public Common::TOCInfo
#endif /* ! defined(SWIG) */
    {
    public:
      typedef FrVect					data_value_type;
      typedef Common::Container< data_value_type >	data_type;
      typedef data_type::const_iterator 		const_iterator;
      typedef data_type::iterator 			iterator;

      typedef Common::Container< FrVect >	aux_type;
      typedef aux_type::const_iterator const_aux_iterator;
      typedef aux_type::iterator aux_iterator;

      typedef Common::Container< FrTable >	table_type;
      typedef table_type::const_iterator const_table_iterator;
      typedef table_type::iterator table_iterator;

      typedef Common::Container< FrHistory >	history_type;
      typedef history_type::const_iterator const_history_iterator;
      typedef history_type::iterator history_iterator;

      typedef General::SharedPtr< FrProcData > subset_ret_type;

      typedef std::string			name_type;
      typedef std::string			comment_type;
      typedef FR_PROC_DATA_TIME_OFFSET_TYPE	timeOffset_type;
      typedef FR_PROC_DATA_T_RANGE_TYPE		tRange_type;
      typedef FR_PROC_DATA_F_SHIFT_TYPE		fShift_type;
      typedef FR_PROC_DATA_PHASE_TYPE		phase_type;
      typedef FR_PROC_DATA_F_RANGE_TYPE		fRange_type;
      typedef FR_PROC_DATA_BW_TYPE		BW_type;

      //-----------------------------------------------------------------
      /// \brief Default constructor
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrProcData( );

      //-----------------------------------------------------------------
      /// \brief Copy Constructor
      ///
      /// \param[in] Source
      ///     The object from which to copy the information.
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrProcData( const FrProcData& Source );

      //-----------------------------------------------------------------
      /// \brief Constructor
      ///
      /// \param[in] Name
      ///     Data or channel name
      /// \param[in] Comment
      ///     Comment 
      /// \param[in] Type
      ///     Type of data object.
      /// \param[in] SubType
      ///     Subtype of f-Series.
      /// \param[in] TimeOffset
      ///     Ofset of 1st sample relative to the frame start time.
      /// \param[in] TRange
      ///     Duration of sampled data.
      /// \param[in] FShift
      ///     The frequency in the original data that corresponds to
      ///     zero Hz in the heterodyned series.
      /// \param[in] Phase
      ///     Phase of heterodyning signal at start of dataset.
      /// \param[in] FRange
      ///     Frequency range.
      /// \param[in] BW
      ///     Resolution bandwidth.
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrProcData( const std::string& Name,
		  const std::string& Comment,
		  type_type Type,
		  subType_type SubType,
		  timeOffset_type TimeOffset,
		  tRange_type TRange,
		  fShift_type FShift,
		  phase_type Phase,
		  fRange_type FRange,
		  BW_type BW );

      //-----------------------------------------------------------------
      /// \brief Constructor
      ///
      /// \param[in] Name
      ///     Data or channel name
      /// \param[in] SampleRate
      ///     Sample rate of data.
      /// \param[in] TimeOffset
      ///     Ofset of 1st sample relative to the frame start time.
      /// \param[in] FShift
      ///     The frequency in the original data that corresponds to
      ///     zero Hz in the heterodyned series.
      /// \param[in] Phase
      ///     Phase of heterodyning signal at start of dataset.
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrProcData( const std::string& Name,
		  REAL_8 SampleRate,
		  const GPSTime& TimeOffset,
		  REAL_8 FShift,
		  REAL_4 Phase );

      //-----------------------------------------------------------------
      /// \brief Promotion Constructor
      ///
      /// \param[in] Source
      ///     An FrProcData structure from a previous frame specification.
      /// \param[in] Stream
      ///     The stream from which the earliest version of the
      ///     FrProcData structure was read.
      ///
      /// \return
      ///     A new instance of an FrProcData object
      //-----------------------------------------------------------------
      FrProcData( Previous::FrProcData& Source,
		  istream_type* Stream );

      //-----------------------------------------------------------------
      /// \brief Destructor
      //-----------------------------------------------------------------
      ~FrProcData( );

      //-----------------------------------------------------------------
      /// \brief Promote previous version of object to this version
      ///
      /// \param[in] Source
      ///     The frame specification version of the object.
      /// \param[in] Obj
      ///     The previous version of the object.
      /// \param[in] Stream
      ///     Stream from which the origional object was created.
      ///
      /// \return
      ///     A new instance of FrameH based on Obj
      //-----------------------------------------------------------------
      static promote_ret_type
      Promote( INT_2U Source,
	       promote_arg_type Obj,
	       istream_type* Stream );

      //-----------------------------------------------------------------
      /// \brief The name structure name of this object.
      ///
      /// \return
      ///     The name of the structure as specified by the frame
      ///     specification.
      //-----------------------------------------------------------------
      static const char* StructName( );

      //-----------------------------------------------------------------
      /// \brief The description of structure
      ///
      /// \return
      ///     A Description object which describes this structure as
      ///     specified by the frame specification.
      //-----------------------------------------------------------------
      static const Common::Description* StructDescription( );

      //-----------------------------------------------------------------
      /// \brief equality operator
      ///
      /// \param[in] RHS
      ///     The FrProcData object to be compared.
      ///
      /// \return
      ///     The value true is returned if this object is equivelent
      ///     to the RHS instance; false otherwise.
      //-----------------------------------------------------------------
      bool operator==( const FrProcData& RHS ) const;

      //-----------------------------------------------------------------
      /// \brief inequality operator
      ///
      /// \param[in] RHS
      ///     The FrProcData object to be compared.
      ///
      /// \return
      ///     The value true is returned if this object is not equivelent
      ///     to the RHS instance; false otherwise.
      //-----------------------------------------------------------------
      bool operator!=( const FrProcData& RHS ) const;

      //-----------------------------------------------------------------
      /// \brief equality operator for abstract data type
      ///
      /// \param[in] Obj
      ///     The object to be compared.
      ///
      /// \return
      ///     The value true is returned if this object is equivelent
      ///     to the RHS instance; false otherwise.
      //-----------------------------------------------------------------
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the frequency range.
      ///
      /// \return
      ///     The frequency range.
      //-----------------------------------------------------------------
      fRange_type GetFRange( ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the time range.
      ///
      /// \return
      ///     The time range.
      //-----------------------------------------------------------------
      tRange_type GetTRange( ) const;

      //-----------------------------------------------------------------
      /// \brief Return the name associate with the FrProcData structure.
      ///
      /// \return
      ///     The name associated with the FrAdcData structure
      //-----------------------------------------------------------------
      const std::string& GetNameSlow( ) const;


      //-----------------------------------------------------------------
      /// \brief Identifier for vector for auxilliary data.
      ///
      /// \return
      ///     A constant reference to the auxilliary data.
      //-----------------------------------------------------------------
      const aux_type& RefAux( ) const;

      //-----------------------------------------------------------------
      /// \brief Identifier for vector for auxilliary data.
      ///
      /// \return
      ///     A reference to the auxilliary data.
      //-----------------------------------------------------------------
      aux_type& RefAux( );

      //-----------------------------------------------------------------
      /// \brief Identifier for vector for data vector.
      ///
      /// \return
      ///     A constant reference to the data vector.
      //-----------------------------------------------------------------
      const data_type& RefData( ) const;

      //-----------------------------------------------------------------
      /// \brief Identifier for vector for data vector.
      ///
      /// \return
      ///     A reference to the data vector.
      //-----------------------------------------------------------------
      data_type& RefData( );

      //-----------------------------------------------------------------
      /// \brief Identifier for parameter table.
      ///
      /// \return
      ///     A constant reference to the parameter table.
      //-----------------------------------------------------------------
      const table_type& RefTable( ) const;

      //-----------------------------------------------------------------
      /// \brief Identifier for parameter table.
      ///
      /// \return
      ///     A reference to the parameter table.
      //-----------------------------------------------------------------
      table_type& RefTable( );

      //-----------------------------------------------------------------
      /// \brief Identifier for channel history.
      ///
      /// \return
      ///     A constant reference to the channel history.
      //-----------------------------------------------------------------
      const history_type& RefHistory( ) const;

      //-----------------------------------------------------------------
      /// \brief Identifier for channel history.
      ///
      /// \return
      ///     A reference to the channel history.
      //-----------------------------------------------------------------
      history_type& RefHistory( );
   
      //-----------------------------------------------------------------
      /// \brief Set the comment associated with the FrAdcData structure.
      //-----------------------------------------------------------------
      void SetComment( const comment_type& comment );

      //-----------------------------------------------------------------
      /// \brief Establish the value of the timeOffset field.
      ///
      /// \param[in] time
      ///     The desired value of the timeOffset field.
      //-----------------------------------------------------------------
      void SetTimeOffset( timeOffset_type time );

      //-----------------------------------------------------------------
      /// \brief Establish the time range.
      ///
      /// \param[in] TRange
      ///     The time range.
      //-----------------------------------------------------------------
      void SetTRange( tRange_type TRange );

      //-----------------------------------------------------------------
      /// \brief Merge with another FrProcData
      ///
      /// \param[in] RHS
      ///     The source of the information to append to this FrProcData
      ///     structure.
      ///
      /// \return
      ///     A reference to this object
      //-----------------------------------------------------------------
      FrProcData& Merge( const FrProcData& RHS );

      //-----------------------------------------------------------------
      /// \brief Request a subset of the data.
      ///
      /// \param[in] Offset
      ///     The number of seconds from the start of the FrProcData
      ///     object from which to start extracting data.
      /// \param[in] Dt
      ///     The number of seconds of data in the resulting FrProcData
      ///     strucutre.
      ///
      /// \return
      ///     A FrProcData structure which contains Dt seconds of data
      ///     starting at Offset of this FrProcData structure.
      //-----------------------------------------------------------------
      subset_ret_type
      Subset( REAL_8 Offset, REAL_8 Dt ) const;

      //-----------------------------------------------------------------
      /// \brief The name structure name of this object.
      ///
      /// \return
      ///     The name of the structure as specified by the frame
      ///     specification.
      //-----------------------------------------------------------------
      virtual const char* ObjectStructName( ) const;

      //-----------------------------------------------------------------
      /// \brief Answer requests for information as needed by TOC
      ///
      /// \param[in] InfoClass
      ///     Class of information being requested.
      //-----------------------------------------------------------------
      virtual void TOCQuery( int InfoClass, ... ) const;

      //-----------------------------------------------------------------
      /// \brief Convert numeric types to string representation
      ///
      /// \param[in] Type
      ///     Numeric type.
      ///
      /// \return
      ///     The string representation of Type.
      //-----------------------------------------------------------------
      static const std::string& IDTypeToString( INT_2U Type );

      //-----------------------------------------------------------------
      /// \brief Convert numeric sub-types to string representation
      ///
      /// \param[in] Type
      ///     Numeric type.
      /// \param[in] SubType
      ///     Numeric sub-type.
      ///
      /// \return
      ///     The string representation of SubType of Type.
      //-----------------------------------------------------------------
      static const std::string& IDSubTypeToString( INT_2U Type,
                                                   INT_2U SubType );

    protected:
      //-----------------------------------------------------------------
      /// \brief Down grade an object
      ///
      /// \param[in] Target
      ///     The version of the frame specification to demote too.
      /// \param[in] Obj
      ///     The version of the object to demote.
      /// \param[in] Stream
      ///     The input stream from which the original object was read.
      ///
      /// \return
      ///     An object of the previous generation.
      //-----------------------------------------------------------------
      virtual demote_ret_type
      demote( INT_2U Target,
	      demote_arg_type Obj,
	      istream_type* Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Upgrade an object
      ///
      /// \param[in] Target
      ///     The version of the frame specification to promote too.
      /// \param[in] Obj
      ///     The object to be promoted.
      /// \param[in] Stream
      ///     The input stream from which the original object was read.
      ///
      /// \return
      ///     An object of this class.
      //-----------------------------------------------------------------
      virtual promote_ret_type
      promote( INT_2U Target,
	       promote_arg_type Obj,
	       istream_type* Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Number of bytes needed to write this structure
      ///
      /// \param[in] Stream
      ///     The stream from which to the object is being read or
      ///     written.
      ///
      /// \return
      ///     The number of bytes need to read or write this object.
      //-----------------------------------------------------------------
      cmn_streamsize_type pBytes( const Common::StreamBase& Stream) const;

      virtual FrProcData* pCreate( istream_type& Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Virtual constructor
      ///
      /// \param[in] Stream
      ///     The input stream from where the object is being read.
      //-----------------------------------------------------------------
      virtual void pWrite( ostream_type& Stream ) const;

    private:
      //-----------------------------------------------------------------
      /// \brief Object with checksum data.
      //-----------------------------------------------------------------
      typedef ObjectWithChecksum< Common::FrameSpec::Object > object_type;

      using Common::FrameSpec::Object::Create;

      //-----------------------------------------------------------------
      /// \brief structure to contain all PTR_STRUCT elements
      //-----------------------------------------------------------------
      struct fr_proc_data_ref_type
      {
	
	//---------------------------------------------------------------
	/// Data vector.
	/// The data vector for single dimensional types (t-Series and
	/// f-Series) must have a single dimension.
	//---------------------------------------------------------------
	data_type		data;
	//---------------------------------------------------------------
	/// Auxiliary data.
	/// Its use is generic.
	//---------------------------------------------------------------
	aux_type		aux;
	//---------------------------------------------------------------
	/// Parameter table.
	//---------------------------------------------------------------
	table_type		table;
	//---------------------------------------------------------------
	/// Channel history.
	/// \note
	///     The first FrHistory should describe the processing used
	///     to build this FrProcData channel.
	///     Its name variable should be the FrProcData name.
	///     If the channel(s) used to produce this FrProcData have
	///     prior FrHistory structures,
	///     then in order to not lose this history,
	///     these structures should be copied to the FrHistory
	///     linked list.
	//---------------------------------------------------------------
	history_type		history;

	//---------------------------------------------------------------
	/// \brief Default Constructor
	//---------------------------------------------------------------
        fr_proc_data_ref_type( );

	//---------------------------------------------------------------
	/// \brief Stream Constructor
	///
	/// \param[in] Stream
	///     The stream from which the object is being read.
	///
	/// \return
	///    A new instance of this object.
	//---------------------------------------------------------------
        fr_proc_data_ref_type( Common::IStream& Stream );

	//---------------------------------------------------------------
	/// \brief Copy non-data components of the structure.
	///
	/// \param[in] Source
	///     The source from which to copy the data.
	//---------------------------------------------------------------
	void copy_core( const fr_proc_data_ref_type& Source );

	//---------------------------------------------------------------
	/// \brief equality operator
	///
	/// \param[in] RHS
	///     The fr_proc_data_ref_type object to be compared.
	///
	/// \return
	///     The value true is returned if this object is equivelent
	///     to the RHS instance; false otherwise.
	//---------------------------------------------------------------
	bool operator==( const fr_proc_data_ref_type& RHS ) const;
      };

      //-----------------------------------------------------------------
      /// Storage of PTR_STRUCT elements
      //-----------------------------------------------------------------
      fr_proc_data_ref_type	m_refs;

      //-----------------------------------------------------------------
      /// True if FrProcData has been syncronized with its associated
      /// FrVect data structure.
      //-----------------------------------------------------------------
      bool			m_synced_with_vector;

      //-----------------------------------------------------------------
      /// \brief Stream Constructor
      ///
      /// \param[in] Stream
      ///     The stream from which the object is being read.
      ///
      /// \return
      ///    A new instance of this object.
      //-----------------------------------------------------------------
      FrProcData( istream_type& Stream );

      //-----------------------------------------------------------------
      /// \brief Copy core data.
      ///
      /// \param[in] Source
      ///     The source from which to copy the data
      //-----------------------------------------------------------------
      void copy_core( const FrProcData& Source );

      //-----------------------------------------------------------------
      /// \brief Retrieve a time series subset.
      ///
      /// \param[in] Offset
      ///     The number of seconds from the start of the FrAdcData
      ///     object from which to start extracting data.
      /// \param[in] Dt
      ///     The number of seconds of data in the resulting FrAdcData
      ///     strucutre.
      ///
      /// \return
      ///     A FrProcData structure which contains Dt seconds of data
      ///     starting at Offset of this FrProcData structure.
      //-----------------------------------------------------------------
      subset_ret_type sub_time_series( REAL_8 Offset,
				       REAL_8 Dt ) const;

      //-----------------------------------------------------------------
      /// \brief Syncronize the FrProcData with the FrVect data.
      //-----------------------------------------------------------------
      void      sync_with_vector( );
    };

    inline void FrProcData::fr_proc_data_ref_type::
    copy_core( const fr_proc_data_ref_type& Source )
    {
      aux = Source.aux;
      table = Source.table;
      history = Source.history;
    }

    inline void FrProcData::
    SetComment( const comment_type& comment )
    {
      m_data.comment = comment;
    } 

    inline bool FrProcData::fr_proc_data_ref_type::
    operator==( const fr_proc_data_ref_type& RHS ) const
    {
      return ( ( this == &RHS )
	       || ( ( data == RHS.data )
		    && ( aux == RHS.aux )
		    && ( table == RHS.table )
		    && ( history == RHS.history ) ) );
    }

    const INT_2U FR_PROC_DATA_DEFAULT_PROC_DATA_TYPE =
      FrProcData::UNKNOWN_TYPE;
    const INT_2U FR_PROC_DATA_DEFAULT_SUB_TYPE = FrProcData::UNKNOWN_SUB_TYPE;
    const REAL_8 FR_PROC_DATA_DEFAULT_TRANGE = 0.0;
    const REAL_8 FR_PROC_DATA_DEFAULT_FSHIFT = 0.0;
    const REAL_4 FR_PROC_DATA_DEFAULT_PHASE = 0.0;
    const REAL_8 FR_PROC_DATA_DEFAULT_FRANGE = 0.0;
    const REAL_8 FR_PROC_DATA_DEFAULT_BW = 0.0;

    inline FrProcData::promote_ret_type FrProcData::
    Promote( INT_2U Source,
	     promote_arg_type Obj,
	     istream_type* Stream )
    {
      return Object::PromoteObject< Previous::FrProcData, FrProcData >
	( DATA_FORMAT_VERSION, Source, Obj, Stream );
    }

    inline const char* FrProcData::
    StructName( )
    {
      static const char* name = "FrProcData";
      return name;
    }

    inline REAL_8 FrProcData::
    GetFRange( ) const
    {
      if ( ! m_synced_with_vector )
      {
	const_cast< FrProcData* >( this )->sync_with_vector( );
      }
      return FrProcDataData::GetFRange( );
    }

    inline FrProcData::tRange_type FrProcData::
    GetTRange( ) const
    {
      if ( ! m_synced_with_vector )
      {
	const_cast< FrProcData* >( this )->sync_with_vector( );
      }
      return FrProcDataData::GetTRange( );
    }

    inline const FrProcData::aux_type& FrProcData::
    RefAux( ) const
    {
      return m_refs.aux;
    }

    inline FrProcData::aux_type& FrProcData::
    RefAux( )
    {
      return m_refs.aux;
    }

    inline const FrProcData::data_type& FrProcData::
    RefData( ) const
    {
      return m_refs.data;
    }

    inline FrProcData::data_type& FrProcData::
    RefData( )
    {
      return m_refs.data;
    }

    inline const FrProcData::history_type& FrProcData::
    RefHistory( ) const
    {
      return m_refs.history;
    }

    inline FrProcData::history_type& FrProcData::
    RefHistory( )
    {
      return m_refs.history;
    }

    inline const FrProcData::table_type& FrProcData::
    RefTable( ) const
    {
      return m_refs.table;
    }

    inline FrProcData::table_type& FrProcData::
    RefTable( )
    {
      return m_refs.table;
    }

    inline void FrProcData::
    SetTimeOffset( timeOffset_type time )
    {
      m_data.timeOffset = time;
    }

    inline void FrProcData::
    SetTRange( tRange_type TRange )
    {
      m_data.tRange = TRange;
    }

    inline bool FrProcData::
    operator==( const FrProcData& RHS ) const
    {
      return ( ( this == &RHS )
	       || ( ( m_data == RHS.m_data )
		    && ( m_refs == RHS.m_refs ) ) );
    }

    inline bool FrProcData::
    operator!=( const FrProcData& RHS ) const
    {
      return ! ( *this == RHS );
    }
  } // namespace - Version_8
} // namespace FrameCPP

#endif /* defined(__cplusplus) */

#endif /* FrameCPP_VERSION_8_FrProcData_HH */
